/*
    HwAccess.cpp - layer to separate chip register details from rc800 program logic
    Copyright (C) 2007  Jeroni Paul

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/

#include <windows.h>
#include "HardwareDriver.h"
#include "PCICard.h"
#include "HwAccess.h"

// Constants:
// Bt8xx IDs
#define PCI_VENDOR_ID_BROOKTREE 0x109e 
#define PCI_DEVICE_ID_BT848     0x350	
#define PCI_DEVICE_ID_BT849     0x351
#define PCI_DEVICE_ID_BT878     0x36e
#define PCI_DEVICE_ID_BT879     0x36f
// Register address
#define MO_GPIO_8x8             0x200
#define MO_GPOE_8x8             0x118
#define MO_PINMUX_IO_8x8        0x4

// Cx2388x IDs
#define PCI_VENDOR_ID_CONEXANT  0x14f1
#define PCI_DEVICE_ID_BT880     0x8800
// Register address
#define MO_GP0_IO           0x350010 // {32}RW* GPIOoutput enablesdata I/O
#define MO_GP1_IO           0x350014 // {32}RW* GPIOoutput enablesdata I/O
#define MO_GP2_IO           0x350018 // {32}RW* GPIOoutput enablesdata I/O
#define MO_GP3_IO           0x35001C // {32}RW* GPIO Mode/Ctrloutput enables
#define MO_GPIO_88x         0x350020 // {32}RW* GPIO I2C Ctrldata I/O
#define MO_GPOE_88x         0x350024 // {32}RW  GPIO I2C Ctrloutput enables
#define MO_PINMUX_IO_88x    0x35C044 // {8}RW  Pin Mux Control

// SAA713x IDs
#define PCI_VENDOR_ID_PHILIPS 0x1131
#define PCI_DEVICE_ID_SAA7130 0x7130
#define PCI_DEVICE_ID_SAA7133 0x7133
#define PCI_DEVICE_ID_SAA7134 0x7134
#define PCI_DEVICE_ID_SAA7146 0x7146
// Register address
#define MO_GPIO_GPMODE        0x1B0
#define MO_GPIO_RESCAN        0x1B3
#define MO_GPIO_GPSTATUS      0x1B4
#define MO_VP_EN_VIDEO        0x193
#define MO_TS_S_EN            0x1A2


// Supported cards
const WORD ids[9][2]={
	{ PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT848 },
	{ PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT849 },
	{ PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT878 },
	{ PCI_VENDOR_ID_BROOKTREE, PCI_DEVICE_ID_BT879 },
	{ PCI_VENDOR_ID_CONEXANT , PCI_DEVICE_ID_BT880 },
	{ PCI_VENDOR_ID_PHILIPS  , PCI_DEVICE_ID_SAA7130 },
	{ PCI_VENDOR_ID_PHILIPS  , PCI_DEVICE_ID_SAA7133 },
	{ PCI_VENDOR_ID_PHILIPS  , PCI_DEVICE_ID_SAA7134 },
	{ PCI_VENDOR_ID_PHILIPS  , PCI_DEVICE_ID_SAA7146 }
};

const char names[9][8]={
	"BT848", "BT849", "BT878", "BT879",	"CX2388X",
	"SAA7130", "SAA7133", "SAA7134", "SAA7146" };

// HardwareDriver and PCICard classes
CHardwareDriver hwDriver;
CPCICard hwCard(&hwDriver);


// Functions
CHwAccess::CHwAccess()
{
	activeCard.dwDeviceID = 0;
	activeCard.dwVendorID = 0;
	activeCard.dwDeviceIndex = 0;
	m_wmask = m_initcard = m_mask = 0;
}

CHwAccess::~CHwAccess()
{
	SetGPOE(m_wmask, m_initgpoe);
}

bool CHwAccess::LoadDriver()
{
	return hwDriver.LoadDriver() == TRUE;
}

int  CHwAccess::FindCards(TCard cards[], int maxcards)
{
	int card = 0, ind_id = 0, dev_id = 0;
	while(card < maxcards)
	{
		if(hwDriver.DoesThisPCICardExist(ids[ind_id][0], ids[ind_id][1], dev_id, m_subsystem))
		{
			cards[card].dwVendorID = ids[ind_id][0];
			cards[card].dwDeviceID = ids[ind_id][1];
			cards[card].description = (char*)names[ind_id];
			cards[card++].dwDeviceIndex = dev_id++;
		}
		else
		{
			dev_id = 0;
			if(++ind_id == 9) break;
		}
	}
	return card;
}

bool CHwAccess::UseCard(const TCard *card)
{
	BOOL r = FALSE;

	if(activeCard.dwVendorID)
	{
		SetChipInit(0, 0);
		SetChipACK(0, 0, m_ACKtime);
		activeCard.dwDeviceID = 0;
		activeCard.dwVendorID = 0;
	}

	if(card->dwVendorID)
		r = hwCard.OpenPCICard(card->dwVendorID, card->dwDeviceID, card->dwDeviceIndex);

	if(r == TRUE)
	{
		activeCard = *card;
		m_initgpoe = GetGPOE();
		GetUsefulBits();
		return true;
	}
	return false;
}


DWORD CHwAccess::GetGPOE()
{
	switch(activeCard.dwVendorID)
	{
	case PCI_VENDOR_ID_BROOKTREE:
		return hwCard.ReadDword(MO_GPOE_8x8);

	case PCI_VENDOR_ID_CONEXANT:
		if(hwCard.ReadByte(MO_GP3_IO+3) & 0x1)
			return hwCard.ReadDword(MO_GP1_IO);
		else
			return hwCard.ReadByte(MO_GP0_IO+1)
				|  (hwCard.ReadByte(MO_GP1_IO+1) << 8)
				|  (hwCard.ReadByte(MO_GP2_IO+1) << 16);

	case PCI_VENDOR_ID_PHILIPS:
		return hwCard.ReadDword(MO_GPIO_GPMODE);
	}

	return 0xFFFFFFFF;
}

DWORD CHwAccess::GetGPDATA()
{
	switch(activeCard.dwVendorID)
	{
	case PCI_VENDOR_ID_BROOKTREE:
		return hwCard.ReadDword(MO_GPIO_8x8);

	case PCI_VENDOR_ID_CONEXANT:
		if(hwCard.ReadByte(MO_GP3_IO+3) & 0x1)
			return hwCard.ReadDword(MO_GP0_IO);
		else
		{
			DWORD data = 0;
			if(m_maskFF)
				data = hwCard.ReadByte(MO_GP0_IO);
			if(m_maskFF00)
				data |= hwCard.ReadByte(MO_GP1_IO) << 8;
			if(m_maskFF0000)
				data |= hwCard.ReadByte(MO_GP2_IO) << 16;
			return data;
		}

	case PCI_VENDOR_ID_PHILIPS:
		BYTE rescan = hwCard.ReadByte(MO_GPIO_RESCAN);
		hwCard.WriteByte(MO_GPIO_RESCAN, (~rescan & 0x80) | (rescan & 0x7F));
		hwCard.WriteByte(MO_GPIO_RESCAN, rescan);
		return hwCard.ReadDword(MO_GPIO_GPSTATUS);
	}
	return 0;
}

DWORD CHwAccess::GetUsefulBits()
{
	DWORD data = GetGPOE();
	BYTE pinmux;
	switch(activeCard.dwVendorID)
	{
	case PCI_VENDOR_ID_BROOKTREE:
		data |= 0xFF000000;
		break;

	case PCI_VENDOR_ID_CONEXANT:
		// Set pins with assigned functions
		pinmux = hwCard.ReadByte(MO_PINMUX_IO_88x);
		data |= 0xFF000000;
		if(pinmux & 0x80) data |= 0xFE0000; // MPEG_PAR_EN
		if(pinmux & 0x40) data |= 0x3F0000; // VIP8_MODE
		if(pinmux & 0x20) data |= 0x30000;  // VIP4_MODE
		if(pinmux & 0x10) data |= 0x7F;     // VIP8_MODE
		if(pinmux & 0x8)  data |= 0xFF00;   // VID8_OUT_EN
		if(pinmux & 0x4)  data |= 0xC0FF00; // VID10_OUT_EN
		if(pinmux & 0x2)  data |= 0xFF;     // VID_IN_EN
		if(pinmux & 0x1)  data |= 0x800000; // HCS1_EN
		break;

	case PCI_VENDOR_ID_PHILIPS:
		if(hwCard.ReadByte(MO_VP_EN_VIDEO) & 0x20)
			data |= 0x1C6FF00;
		if(hwCard.ReadByte(MO_VP_EN_VIDEO) & 0x40)
			data |= 0x1C600FF;
		if(hwCard.ReadByte(MO_TS_S_EN) & 0x80)
			data |= 0x7900FF;
		data |= 0xF1000000;
		break;
	}
	
	m_mask = ~data;
	m_maskFF = (m_mask & 0xFF) != 0;
	m_maskFF00 = (m_mask & 0xFF00) != 0;
	m_maskFF0000 = (m_mask & 0xFF0000) != 0;
	return m_mask;
}

void CHwAccess::SetChipInit(DWORD wmask, DWORD wvalue)
{
	SetGPOE(m_initcard, m_initgpoe | m_wmask);
	m_initcard = wmask;
	SetGPDATA(wmask, (GetGPDATA() & ~wmask) | (wvalue & wmask));
	SetGPOE(wmask, wmask | m_initgpoe | m_wmask);
}

void CHwAccess::SetChipACK(DWORD wmask, DWORD wvalue, DWORD ACKtime)
{ 
	// Restore original GPOE
	SetGPOE(m_wmask, m_initgpoe | m_initcard);
	// Set new one
	m_wmask = wmask;
	m_wvalue = wvalue; 
	m_ACKtime = ACKtime;
	SetGPOE(wmask, wmask | m_initgpoe | m_initcard);
}

void CHwAccess::SetGPOE(DWORD mask, DWORD value)
{ 
	switch(activeCard.dwVendorID)
	{
	case PCI_VENDOR_ID_BROOKTREE:
		if(mask & 0xFF)
			hwCard.WriteByte(MO_GPOE_8x8, (BYTE)value & 0xFF);
		if(mask & 0xFF00)
			hwCard.WriteByte(MO_GPOE_8x8+1, (BYTE)(value >> 8) & 0xFF);
		if(mask & 0xFF0000)
			hwCard.WriteByte(MO_GPOE_8x8+2, (BYTE)(value >> 16) & 0xFF);
		return;

	case PCI_VENDOR_ID_CONEXANT:
		if(hwCard.ReadByte(MO_GP3_IO+3) & 0x1)
		{
			if(mask & 0xFF)
				hwCard.WriteByte(MO_GP1_IO, (BYTE)value & 0xFF);
			if(mask & 0xFF00)
				hwCard.WriteByte(MO_GP1_IO+1, (BYTE)(value >> 8) & 0xFF);
			if(mask & 0xFF0000)
				hwCard.WriteByte(MO_GP1_IO+2, (BYTE)(value >> 16) & 0xFF);
		}
		else
		{
			if(mask & 0xFF)
				hwCard.WriteByte(MO_GP0_IO+1, (BYTE)value & 0xFF);
			if(mask & 0xFF00)
				hwCard.WriteByte(MO_GP1_IO+1, (BYTE)(value >> 8) & 0xFF);
			if(mask & 0xFF0000)
				hwCard.WriteByte(MO_GP2_IO+1, (BYTE)(value >> 16) & 0xFF);
		}
		return;

	case PCI_VENDOR_ID_PHILIPS:
		if(mask & 0xFF)
			hwCard.WriteByte(MO_GPIO_GPMODE, (BYTE)value & 0xFF);
		if(mask & 0xFF00)
			hwCard.WriteByte(MO_GPIO_GPMODE+1, (BYTE)(value >> 8) & 0xFF);
		if(mask & 0xFF0000)
			hwCard.WriteByte(MO_GPIO_GPMODE+2, (BYTE)(value >> 16) & 0xFF);
		if(mask & 0x0E000000)
			hwCard.WriteByte(MO_GPIO_GPMODE+3, (BYTE)(value >> 24) & 0x0E);
		return;
	}
}


void CHwAccess::SetGPDATA(DWORD mask, DWORD value)
{ 
	switch(activeCard.dwVendorID)
	{
	case PCI_VENDOR_ID_BROOKTREE:
		if(mask & 0xFF)
			hwCard.WriteByte(MO_GPIO_8x8, (BYTE)value & 0xFF);
		if(mask & 0xFF00)
			hwCard.WriteByte(MO_GPIO_8x8+1, (BYTE)(value >> 8) & 0xFF);
		if(mask & 0xFF0000)
			hwCard.WriteByte(MO_GPIO_8x8+2, (BYTE)(value >> 16) & 0xFF);
		return;

	case PCI_VENDOR_ID_CONEXANT:
		if(hwCard.ReadByte(MO_GP3_IO+3) & 0x1)
		{
			if(mask & 0xFF)
				hwCard.WriteByte(MO_GP0_IO, (BYTE)value & 0xFF);
			if(mask & 0xFF00)
				hwCard.WriteByte(MO_GP0_IO+1, (BYTE)(value >> 8) & 0xFF);
			if(mask & 0xFF0000)
				hwCard.WriteByte(MO_GP0_IO+2, (BYTE)(value >> 16) & 0xFF);
		}
		else
		{
			if(mask & 0xFF)
				hwCard.WriteByte(MO_GP0_IO, (BYTE)value & 0xFF);
			if(mask & 0xFF00)
				hwCard.WriteByte(MO_GP1_IO, (BYTE)(value >> 8) & 0xFF);
			if(mask & 0xFF0000)
				hwCard.WriteByte(MO_GP2_IO, (BYTE)(value >> 16) & 0xFF);
		}
		return;

	case PCI_VENDOR_ID_PHILIPS:
		if(mask & 0xFF)
			hwCard.WriteByte(MO_GPIO_GPSTATUS, (BYTE)value & 0xFF);
		if(mask & 0xFF00)
			hwCard.WriteByte(MO_GPIO_GPSTATUS+1, (BYTE)(value >> 8) & 0xFF);
		if(mask & 0xFF0000)
			hwCard.WriteByte(MO_GPIO_GPSTATUS+2, (BYTE)(value >> 16) & 0xFF);
		if(mask & 0x0E000000)
			hwCard.WriteByte(MO_GPIO_GPSTATUS+3, (BYTE)(value >> 24) & 0x0E);
		return;
	}
}

void CHwAccess::ChipACK()
{
	if(m_wmask)
	{
		DWORD data = GetGPDATA();
		SetGPDATA(m_wmask, (~m_wmask & data) | (m_wmask & m_wvalue));
		Sleep(m_ACKtime);
		data = GetGPDATA();
		SetGPDATA(m_wmask, (~m_wmask & data) | (m_wmask & ~m_wvalue));
	}
}
